(* = Atools.ML 9.3.01
   +++ outcommented tests *)

(** eval_ 
    precondition: head_of t is the constant to be evaluated 
                  (had been filtered by get_pair before call) **)
(* eval_const was the reason for introducing exception GO_ON:
   patterns on arg incomparable to other unary functions.

   ugly consequence: 
   eval_... need to have the same result type as get_pair
*)
fun eval_const (thmid:string) _ (t as (Const(op0,t0) $ arg)) (thy:theory) = 
    (case arg of 
       Const(n1,_) =>
	 Some (mk_thmid thmid op0 n1 "", 
	       Trueprop $ (mk_equality (t, false_as_term)))
     | Free(n1,_) => 
	 if (hd o explode) n1 = "#"
	   then Some (mk_thmid thmid op0 n1 "", 
		      Trueprop $ (mk_equality (t, true_as_term)))
	 else Some (mk_thmid thmid op0 n1 "", 
		    Trueprop $ (mk_equality (t, false_as_term)))
     | _ => raise GO_ON)
  | eval_const _ _ _ _ = raise GO_ON; 
  (* necessary for recursion over (t1$t2) in get_pair *)
(* use"Isa99/knowl-base/RatArith.ML";

> val ct = (the o (parse thy)) "#3 is_const";
> calc_thm thy "is'_const" ct;
val it = Some ("is_const_3_","(is_const 3 ) = True  [(is_const 3 ) = True]")
*)

(* fn : string -> string -> term -> string * term 
   TODO clumsy: 2-fold pattern matching
                3 patterns for binops used for equ, too
                  although only the 1st pattern matches ... *)
fun eval_binop (thmid:string) (op_:string) (t as 
	       (Const (op0,t0) $ Free (n1,t1) $ Free(n2,t2))) thy = 
                                                (* binary . n1.n2 *)
  let
    val n1' = int_of_str n1;
    val n2' = int_of_str n2;
    val (T1,T2,Trange) = dest_binop_typ t0;
    val res = calc (strip_thy op0) (n1',n2');
    val rhs = term_of_num Trange res;
    val prop = Trueprop $ (mk_equality (t, rhs))
  in Some (mk_thmid thmid op0 n1 n2, prop) end
  | eval_binop (thmid:string) (op_:string) (t as ( Const(op0,t0) $ 
			   ( Const(op0',t0') $ v $ Free(n1,t1)) $
			   Free(n2,t2))) thy = 
  let                                       (* binary . (v.n1).n2 *)
    val n1' = int_of_str n1;
    val n2' = int_of_str n2;
    val (T1,T2,Trange) = dest_binop_typ t0;
    val res = calc (strip_thy op0) (n1',n2');
    val rhs = var_op_num v op_ t0 T1 res;
    val prop = Trueprop $ (mk_equality (t, rhs))
  in Some (mk_thmid thmid op0 n1 n2, prop) end
  | eval_binop (thmid:string) (op_:string) (t as 
			   ( Const(op0,t0) $ Free(n1,t1) $ 
			    ( Const(op0',t0') $ Free(n2,t2) $ v))) thy =
  let                                      (* binary . n1.(n2.v) *)
    val n1' = int_of_str n1;
    val n2' = int_of_str n2;
    val (T1,T2,Trange) = dest_binop_typ t0;
    val res = calc (strip_thy op0) (n1',n2');
    val rhs = num_op_var v op_ t0 T1 res;
    val prop = Trueprop $ (mk_equality (t, rhs))
  in Some (mk_thmid thmid op0 n1 n2, prop) end
  | eval_binop _ _ _ _ = raise GO_ON; 

fun eval_equ (thmid:string) (op_:string) (t as 
	       (Const (op0,t0) $ Free (n1,t1) $ Free(n2,t2))) thy = 
  if calc_equ (strip_thy op0) ((int_of_str n1),(int_of_str n2))
    then Some (mk_thmid thmid op0 n1 n2, 
	  Trueprop $ (mk_equality (t, true_as_term)))
  else Some (mk_thmid thmid op0 n1 n2,  
	  Trueprop $ (mk_equality (t, false_as_term)))
  | eval_equ _ _ _ _ = raise GO_ON;


fun eval_ident (thmid:string) (op_:string) (t as 
	       (Const (op0,t0) $ t1 $ t2 )) thy = 
  if t1 = t2
    then Some (mk_thmid thmid op0 
	       ("("^(Sign.string_of_term (sign_of thy) t1)^")")
	       ("("^(Sign.string_of_term (sign_of thy) t2)^")"), 
	  Trueprop $ (mk_equality (t, true_as_term)))
  else Some (mk_thmid thmid op0  
	       ("("^(Sign.string_of_term (sign_of thy) t1)^")")
	       ("("^(Sign.string_of_term (sign_of thy) t2)^")"),  
	  Trueprop $ (mk_equality (t, false_as_term)))
  | eval_ident _ _ _ _ = raise GO_ON;
(*
 (string_of_cterm o (cterm_of (sign_of thy)));
*)

val simprls = 
  Rls{preconds = [], rew_ord = ("tless_true",tless_true), 
      rules = [Calc ("ident",eval_ident "#ident_"),
	       Calc ("Var",eval_var "#Var_"),
	       Calc ("Length",eval_Length "#Length_"),
	       Calc ("Nth",eval_Nth "#Nth_"),
	       Calc ("op <",eval_equ "#less_"),
	       Calc ("op <=",eval_equ "#less_equal_"),
	       Calc ("ident",eval_ident "#ident_"),

	       Thm ("if_True",num_str if_True),
	       Thm ("if_False",num_str if_False)
	       ],
      scr = Script ((term_of o the o (parse thy)) 
      "empty_script")
      }:rls;      

